#
#===============================================================================
#
#         FILE:  LogParser.pm
#
#  DESCRIPTION:  This module is used for opening, and parsing the log file
#
#        FILES:  ---
#         BUGS:  ---
#        NOTES:  ---
#       AUTHOR:   (sathiyamoorthy), <sathiyamoorthy@bksys.co.in>
#      COMPANY:  
#      VERSION:  1.0
#      CREATED:  07/11/08 10:05:58 IST
#     REVISION:  ---
#===============================================================================

use strict;
use warnings;

use PQAGlobals;


#===  FUNCTION  ================================================================
#         NAME:  OpenLogFile
#      PURPOSE:  To open the given log file
#   PARAMETERS:  Logfile name with with relative or absolute path
#      RETURNS:  
#      	0 if the file is not openable
#      	FilePointer if the file got opened successfully
#  DESCRIPTION:  
# 		The log file given in the command line will be opened, and the file pointer 
# 		will be returned to the caller of this function.
#       THROWS:  no exceptions
#     COMMENTS:  none
#     SEE ALSO:  n/a
#===============================================================================

sub OpenLogFile {
	my $OPENLOGFILE;
	my $file_open_status;

	# parsing the file name from the command line.,
	my $db_log_file_name = $PQAGlobals::options{"f"};

#---------------------------------------------------------------------------
#  if the file is compressed file, then use appropriate module to open it
#  else use simple open, and use it. ( i will not ask the module PerlIO::gzip 
#  unwantedly, if the file is compressed then only i ll ask ).
#---------------------------------------------------------------------------
	if ( defined $PQAGlobals::options{"z"} )  {
		# following is one way of conditional module usage.,
		eval { 
			require PerlIO::gzip;
		};
		die "install module PerlIO::gzip \n" if $@;
		$file_open_status = open $OPENLOGFILE, "<:gzip", $db_log_file_name;
	} else  {
		# opening log file, if it is a text file.	
		$file_open_status = open $OPENLOGFILE, "<$db_log_file_name";
	}

	# check the file open status, and report the error, and exit.
	if ( ! $file_open_status )  {
		print "Error: Unable to open the Log file $db_log_file_name";
		exit;
	}
	
	# if file opened successfully, return the file handle.,
	return $OPENLOGFILE;
}


#===  FUNCTION  ================================================================
#         NAME:  MakeAsOneLine 
#      PURPOSE:  make the multiple lines into a single line
#   PARAMETERS:  a line from the log file
#      RETURNS:  ????
#  DESCRIPTION:  This function parses the postgres log line, makes continuous lines
#  as single line, and then calls the parsedblog function to parse line by line.
#       THROWS:  no exceptions
#     COMMENTS:  none
#     SEE ALSO:  n/a
#===============================================================================

sub MakeAsOneLine  {
	CheckArg(@_,1);

	my ($PARSEFILE) = shift @_;

	my ($cur_line);
	my $only_once_true_flag=1;
	my $i=0;
	
	# parse the input file line, by line.
	while(($_ = scalar <$PARSEFILE>))  {
	# remove the new line from that line.
		chomp;
	
	# if the line contains the tab as the first character, then that is the "CONTINUATION LINE",
	# so remove the "whitespace character from that".,
		if ( /^\t/ )  {
	# if there is no previous line then print the warning and continue reading the next line.,
			if ( not defined $cur_line )  {
				warn "continuation of no previous line\n";
				#last;
				next;
			}
	# since it is a continuation line, concatenate it with the previous line.,		
			$cur_line.=$_;
		} else {
	# if the line does not contains the tab, then take that line as the current line.,	
		make_as_one_line_lbl :
			if (defined $cur_line) {
				# check what is the current line.
				CheckThisLogLine($cur_line);
			}

			# store the current line in the variable		
			# $cur_line = $old;	
			$cur_line = $_;	
		}

	# last line is not parsed still now, so go to the label and parse the last line, and
	# automatically it will exit., and this flag also became false, 
		if ( eof $PARSEFILE)  {
			if ( $only_once_true_flag )  {
				$only_once_true_flag = 0;
				goto make_as_one_line_lbl;
				}
		}	
	}

	#use Data::Dumper;
	#print Dumper(%PQAGlobals::query_details);
}


sub CheckThisLogLine {
	CheckArg(@_,1);	

	my ( $CurDuration, $Pid );

	# store the current line in the $_	
	local $_ = shift @_;

	# this is a statement line
	if ( /$PQAGlobals::STATEMENT_LINE/i )  {
		# if it contains duration, copy it, and remove the duration
		if ( /$PQAGlobals::DURATION_WITH_STATEMENT/i )  {
			# copy the duration from the log line.
			$CurDuration = $1;
			$CurDuration = $1 * 1000 if $2 eq "sec";

			# this will remove Duration: time (ms|sec) from the Log line
			# substitute the values with the braces., $3 -> send only statement.
			$_=SubstituteQuery($3);
			PopulateHash($_, $CurDuration);
		} elsif ( /$PQAGlobals::MATCH_PID$PQAGlobals::STATEMENT_LINE/i )  {
			# /$PQAGlobals::MATCH_PID$PQAGlobals::STATEMENT_LINE/
			# just matching the PID matches the array indexes also, so match it like pid & statement	
			# line contains the pid, and also the statement.
			$Pid = $1;

			# substitute the values with the braces., $2 -> send only the statement as argument
			$_=SubstituteQuery($2);
			PopulateTmpPidHash($_, $Pid, $CurDuration);
		} else {
			# line contains only statement, no pid, no duration.,
			
			# substitute the values with the braces., $1 -> send only the statement as argument
			$_=SubstituteQuery($1);
			PopulateHash($_);
		}
	} elsif ( /$PQAGlobals::MATCH_PID$PQAGlobals::DURATION/i )  {
			$Pid = $1;
			# if the unit is seconds then multiply it with 1000, and store it as ms.
			( $3 eq "sec" ) ? $CurDuration = $2 * 1000 : $CurDuration = $2;

			# pid, and duration only will be there in the current line, so send query as undef.
			PopulateTmpPidHash(undef, $Pid, $CurDuration);
	} elsif ( /$PQAGlobals::DURATION/i ) {
			# line contains only duration, then store that as duration for the previous query.,
			# if the unit is seconds then multiply it with 1000, and store it as ms.
			( $2 eq "sec" ) ? $CurDuration = $1 * 1000 : $CurDuration = $1;
			PopulateHash($PQAGlobals::previous_query, $CurDuration);
	} elsif ( /$PQAGlobals::ERROR_LINE/ )  {
		# fill the errors in to error hash.
		$PQAGlobals::error_query_details{$1}++;
	} else {
		# if the line does not contains any useful info such as STATEMENT, or DURATION, then discard it.,
		return 1;
	}

	# always the previous query in the global variable
	$PQAGlobals::previous_query=$_;
}	# ----------  end of subroutine CheckThisLogLine  ----------


#===  FUNCTION  ================================================================
#         NAME:  PopulateTmpPidHash
#      PURPOSE:  manipulate temporary hash with the pid, query, and its duration.,
#   PARAMETERS:  
#   		$TmpQuery -> query
#   		$TmpPid -> pid of that query
#   		$TmpDuration -> duration of that query
#      RETURNS:  ????
#  DESCRIPTION:  populate a hash with the pid, and its query., if duration is there
#  			move that to the main hash.
#       THROWS:  no exceptions
#     COMMENTS:  none
#     SEE ALSO:  n/a
#===============================================================================
sub PopulateTmpPidHash {
	CheckArg(@_,3);
	my ( $TmpQuery, $TmpPid, $TmpDuration ) = @_;

#---------------------------------------------------------------------------
# PID handling			
# when PID and statement occurs, store this PID as a key and store the
# query in the value.
# next time when PID and duration comes, store all these values into the main 
# query hash and delete from temppidhash key.
#---------------------------------------------------------------------------

	if ( not defined $TmpQuery )  {
		# if only pid, and duration is there move to the main hash
		if ( not defined $PQAGlobals::TmpHash{$TmpPid} )  {
			warn "duration for no previous query\n";
			return;
		} 

		PopulateHash($PQAGlobals::TmpHash{$TmpPid}, $TmpDuration);
		delete $PQAGlobals::TmpHash{$TmpPid};
	} else {
		# if pid and query is only there populate in the temporary hash
		$PQAGlobals::TmpHash{$TmpPid}=$TmpQuery;
	}
	
	
	return ;
}	# ----------  end of subroutine PopulateTmpPidHash  ----------


#===  FUNCTION  ================================================================
#      PURPOSE:  populate the hash with the statement, pid, and duration.,
#   PARAMETERS:  query, duration, pid (if the duration and pid is availale 
#   			 in that query send those).
#      RETURNS:  
#  DESCRIPTION:  
#       THROWS:  no exceptions
#     COMMENTS:  none
#     SEE ALSO:  n/a
#===============================================================================


sub PopulateHash {

	my ($Duration);

	$_ = shift @_;
	$Duration = shift @_;
	

	# if the query which comes is getting into hash very first
	if ( not defined $PQAGlobals::query_details{$_} )  {
		# assign the min duration as ZERO
		$PQAGlobals::query_details{$_}->[2]=-1;
		# assign the max duration as ZERO
		$PQAGlobals::query_details{$_}->[3]=-1;
		# assign the total duration as ZERO
		$PQAGlobals::query_details{$_}->[4]=0;
	}

	# increment this query's counter. 
	$PQAGlobals::query_details{$_}->[0]++;


	## processing duration.
	## {query} => [counter, flag, min duration, max duration, total duration, [durations], [PID]]
	## here 5th index of an array is duration array.. so for this query whatever the duration
	## comes we have to push that into 5th index. 
	#if ( $Duration != 0){
	if ( defined $Duration ){
		# if duration is there then, 
		# 			1) push that in array, 
		# 			2) add that in total duration.
		push(@{$PQAGlobals::query_details{$_}->[5]},$Duration);
		$PQAGlobals::query_details{$_}->[4]+=$Duration;

		# if the min duration is -1, then the current duration is the max and min duration.,	
 		if ( $PQAGlobals::query_details{$_}->[2]==-1 )  {
			$PQAGlobals::query_details{$_}->[3]=$PQAGlobals::query_details{$_}->[2]=$Duration;
			return;
		};	
			
		## If this Duration is minimum, store this duration as minimum duration.
		if ( $Duration < $PQAGlobals::query_details{$_}->[2] ){
			$PQAGlobals::query_details{$_}->[2]=$Duration;
			return;
		}

		## If this Duration is maximum, store this duration as maximum duration.
		if ( $Duration > $PQAGlobals::query_details{$_}->[3] ){
			$PQAGlobals::query_details{$_}->[3]=$Duration;
		}
	}

	return 1;
}	# ----------  end of subroutine PopulateHash  ----------


#===  FUNCTION  ================================================================
#         NAME:  SubstituteQuery
#      PURPOSE:  substitute the query values with {}
#   PARAMETERS:  nothing
#      RETURNS:  Original query should be affected.,
#  DESCRIPTION: 
#  	1. Remove the space in " = " and make it as "="
#  	2. substitute the values with {} from "a='10'" to "a={}"
#       THROWS:  no exceptions
#     COMMENTS:  none
#     SEE ALSO:  n/a
#===============================================================================

sub SubstituteQuery {
	CheckArg(@_,1);

	# shift the query into $_
	local $_ = shift @_;

	# substitute all the white spaces with single white space, see this is global substitution		
	s/\s+/ /g;

	# the query may be like column = 'value' or column='value'. when storing the
	# query as a hash of the key, we should have query as unique. 
	# so we have to make the query as column='value'.
	# so here substituting <space><operator>+<space to <operator>.
	s/ ?([=<>~!]+) ?/$1/g;

	my $RemoveValue = "'[^']*'";

	## previously it did not match the query like a.public_id=1 
	my $RemoveNumber = '(\W)\d+';
	
	# replace the values with braces {}
	s/$RemoveValue/{}/g;
	s/$RemoveNumber/$1\{\}/g;

	# removing leading and trailing space
	s/^\s+//g;
	s/\s+$//g;

	# make every thing to lower case, # else it will have different keys for same query with different case.
	s/(.*)/\L$1/g;

	#print ">>$_<<\n";
	
	return $_;
}	# ----------  end of subroutine SubstituteQuery  ----------


1;

